# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Graz University of Technology.
#
# invenio-workflows-tugraz is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Models for theses workflow."""

from invenio_db import db
from sqlalchemy import BOOLEAN


class WorkflowThesesMetadata(db.Model):
    """Workflow theses."""

    __tablename__ = "workflows_theses"

    pid = db.Column(db.String(255), primary_key=True)

    cms_id = db.Column(db.Integer)

    imported_in_repo = db.Column(BOOLEAN, default=False)

    archived_in_cms = db.Column(BOOLEAN, default=False)

    created_in_alma = db.Column(BOOLEAN, default=False)

    updated_in_repo = db.Column(BOOLEAN, default=False)

    published_in_cms = db.Column(BOOLEAN, default=False)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Graz University of Technology.
#
# invenio-workflows-tugraz is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Alembic create branch for theses workflow."""


# revision identifiers, used by Alembic.
revision = "9d446c6a77e2"
down_revision = None
branch_labels = ("workflows_theses",)
depends_on = "dbdbc1b19cf2"


def upgrade() -> None:
    """Upgrade database."""


def downgrade() -> None:
    """Downgrade database."""



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
#
# Copyright (C) 2022-2024 Graz University of Technology.
#
# Invenio-Records-Marc21 is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.


"""Blueprint definitions."""

from flask import Blueprint, Flask, g, redirect
from invenio_records_marc21.proxies import current_records_marc21
from werkzeug.wrappers import Response as BaseResponse

from .decorators import pass_record_from_pid


@pass_record_from_pid
def record_from_pid(record: dict, **__: dict) -> BaseResponse:
    """Redirect to record's latest version page."""
    return redirect(record["links"]["self_html"], code=301)


@pass_record_from_pid
def uploads_from_thesesid(record: dict, **__: dict) -> BaseResponse:
    """Upload introduced by theses id."""
    marcid = record["id"]
    current_records_marc21.records_service.edit(g.identity, marcid)
    return redirect(f"/publications/uploads/{marcid}", code=301)


def create_blueprint(_: Flask) -> Blueprint:
    """Register blueprint routes on app."""
    blueprint = Blueprint(
        "invenio_workflows_tugraz_theses",
        __name__,
        url_prefix="/theses",
    )

    blueprint.add_url_rule("/<path:pid_value>", view_func=record_from_pid)
    blueprint.add_url_rule("/<path:pid_value>/uploads", view_func=uploads_from_thesesid)

    return blueprint



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2024 Graz University of Technology.
#
# invenio-workflows-tugraz is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Pytest configuration.

See https://pytest-invenio.readthedocs.io/ for documentation on which test
fixtures are available.
"""


from collections import namedtuple
from collections.abc import Callable

import pytest
from flask import Flask
from invenio_db import InvenioDB
from invenio_records_marc21 import InvenioRecordsMARC21
from invenio_search import InvenioSearch

from invenio_workflows_tugraz import InvenioWorkflowsTugraz


@pytest.fixture(scope="module")
def create_app(instance_path: str) -> Callable:
    """Application factory fixture."""

    def factory(**config: dict) -> Flask:
        app = Flask("testapp", instance_path=instance_path)
        app.config.update(**config)
        InvenioWorkflowsTugraz(app)
        InvenioSearch(app)
        InvenioDB(app)
        InvenioRecordsMARC21(app)
        return app

    return factory


@pytest.fixture()
def embargoed_record_xml() -> str:
    """Embargoed record fixture."""
    return """
    <record xmlns="http://www.loc.gov/MARC21/slim">
          <leader>06215nam a2200505 c 4500</leader>
          <controlfield tag="009">AC11111111</controlfield>
          <datafield ind1="7" ind2=" " tag="971">
            <subfield code="a">gesperrt</subfield>
            <subfield code="b">27.01.2023</subfield>
            <subfield code="c">27.01.2025</subfield>
          </datafield>
    </record>
    """


RunningApp = namedtuple(  # noqa: PYI024
    "RunningApp",
    [
        "app",
        "location",
        "cache",
    ],
)


@pytest.fixture()
def running_app(
    app,  # noqa: ANN001
    location,  # noqa: ANN001
    cache,  # noqa: ANN001
) -> RunningApp:
    """Fixture provides an app with the typically needed db data loaded."""
    return RunningApp(
        app,
        location,
        cache,
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Graz University of Technology.
#
# invenio-workflows-tugraz is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Tasks for theses workflow."""

from datetime import datetime, timezone

from celery import shared_task
from flask import current_app
from invenio_access.permissions import system_identity
from invenio_campusonline import current_campusonline

from ..proxies import current_workflows_tugraz


@shared_task(ignore_result=True)
def status_arch() -> None:
    """Set status to ARCH (archived)."""
    today = datetime.now(tz=timezone.utc).strftime("%Y-%m-%dT%H:%M:%S")

    theses_service = current_workflows_tugraz.theses_service
    cms_service = current_campusonline.campusonline_rest_service

    entries = theses_service.get_ready_to(system_identity, state="archive_in_cms")

    for entry in entries:
        cms_id = entry.cms_id
        pid = entry.pid

        try:
            cms_service.set_status(system_identity, cms_id, "ARCH", today)
            theses_service.set_state(system_identity, pid, state="archived_in_cms")

            msg = "Theses %s has been archived successfully."
            current_app.logger.info(msg, cms_id)
        except RuntimeError as e:
            msg = "Theses %s have been produced error %s on archiving."
            current_app.logger.error(msg, cms_id, str(e))


@shared_task(ignore_result=True)
def status_pub() -> None:
    """Set status to PUB (published)."""
    today = datetime.now(tz=timezone.utc).strftime("%Y-%m-%dT%H:%M:%S")

    theses_service = current_workflows_tugraz.theses_service
    cms_service = current_campusonline.campusonline_rest_service
    entries = theses_service.get_ready_to(system_identity, state="publish_in_cms")

    for entry in entries:
        cms_id = entry.cms_id
        pid = entry.pid

        try:
            cms_service.set_status(system_identity, cms_id, "PUB", today)
            theses_service.set_state(system_identity, pid, state="published_in_cms")

            msg = "Theses %s has been published successfully."
            current_app.logger.info(msg, cms_id)
        except RuntimeError as e:
            msg = "Theses %s have been produced error %s on publishing."
            current_app.logger.error(msg, cms_id, str(e))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Graz University of Technology.
#
# invenio-workflows-tugraz is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""API for theses workflow."""

from __future__ import annotations

from invenio_db import db

from .models import WorkflowThesesMetadata


class WorkflowTheses:
    """Workflow thesis api."""

    model_cls = WorkflowThesesMetadata

    def __init__(self, model: WorkflowThesesMetadata = None) -> None:
        """Construct WorkflowTheses."""
        self.model = model

    @property
    def pid(self) -> str:
        """Get Pid."""
        return self.model.pid

    @property
    def cms_id(self) -> str:
        """Get cms_id."""
        return self.model.cms_id

    @classmethod
    def resolve(cls, id_: str):  # noqa: ANN206
        """Get."""
        model = cls.model_cls.query.filter_by(pid=id_).one_or_none()
        return cls(model=model)

    @classmethod
    def create(cls, id_: str, cms_id: str):  # noqa: ANN206
        """Create."""
        with db.session.begin_nested():
            entry = cls(model=cls.model_cls(pid=id_, cms_id=cms_id))
            db.session.add(entry.model)
        return entry

    def commit(self) -> None:
        """Commit."""
        with db.session.begin_nested():
            db.session.merge(self.model)

    def set_state(self, state: str) -> None:
        """Set archived."""
        if state == "archived_in_cms":
            self.model.archived_in_cms = True
        if state == "imported_in_repo":
            self.model.imported_in_repo = True
        if state == "created_in_alma":
            self.model.created_in_alma = True
        if state == "updated_in_repo":
            self.model.updated_in_repo = True
        if state == "published_in_cms":
            self.model.published_in_cms = True
        db.session.merge(self.model)

    @classmethod
    def get_ready_to(cls, state: str) -> list:
        """Get ready to."""
        entries = []
        if state == "archive_in_cms":
            entries = cls.model_cls.query.filter_by(
                imported_in_repo=True,
                archived_in_cms=False,
            )

        if state == "create_in_alma":
            entries = cls.model_cls.query.filter_by(
                archived_in_cms=True,
                created_in_alma=False,
            )

        if state == "update_in_repo":
            entries = cls.model_cls.query.filter_by(
                created_in_alma=True,
                updated_in_repo=False,
            )

        if state == "publish_in_cms":
            entries = cls.model_cls.query.filter_by(
                updated_in_repo=True,
                published_in_cms=False,
            )

        return [cls(model=entry) for entry in entries.all()]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023-2024 Graz University of Technology.
#
# invenio-workflows-tugraz is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Module test theses."""

from xml.etree.ElementTree import fromstring

from flask import Flask
from flask_principal import Identity
from invenio_access.permissions import system_identity
from invenio_records_marc21.proxies import current_records_marc21
from invenio_records_resources.services.uow import UnitOfWork

from invenio_workflows_tugraz.proxies import current_workflows_tugraz
from invenio_workflows_tugraz.theses.theses import update_func


def test_update_func(
    running_app: Flask,  # noqa: ARG001
    embargoed_record_xml: str,
) -> None:
    """Test update func."""

    class MockRecordItem:
        """Mock RecordItem class."""

        @property
        def data(self) -> dict:
            """Mock data."""
            return {
                "metadata": {
                    "leader": "",
                    "fields": {
                        "971": [
                            {
                                "ind1": "7",
                                "ind2": " ",
                                "subfields": {
                                    "a": [
                                        "gesperrt",
                                    ],
                                    "b": [
                                        "27.01.2023",
                                    ],
                                    "c": [
                                        "27.01.2025",
                                    ],
                                },
                            },
                        ],
                    },
                },
                "access": {
                    "record": "restricted",
                    "files": "restricted",
                },
            }

    class MockRecordsService:
        """Mock RecordsService class."""

        def read_draft(self, id_: str, identity: Identity) -> dict:  # noqa: ARG002
            """Mock read_draft."""
            return MockRecordItem()

        def edit(self, id_: str, identity: Identity) -> None:
            """Mock edit."""

        def update_draft(
            self,
            id_: str,  # noqa: ARG002
            identity: Identity,  # noqa: ARG002
            data: dict,
        ) -> None:
            """Mock update draft."""
            assert data["access"]["record"] == "public"
            assert data["access"]["files"] == "restricted"

        def publish(self, id_: str, identity: Identity) -> None:
            """Mock publish."""

    class MockAlmaService:
        """Mock AlmaService class."""

        def get_record(self, cms_id: str, search_key: str) -> list[str]:  # noqa: ARG002
            doc = fromstring(embargoed_record_xml)  # noqa: S314
            return [doc]

    class MockThesesService:
        """Mock ThesesService class."""

        def set_state(
            self,
            identity: Identity,
            id_: str,
            state: str,
            uow: UnitOfWork = None,
        ) -> None:
            """Mock set state."""

    current_records_marc21.records_service = MockRecordsService()
    current_workflows_tugraz.theses_service = MockThesesService()
    alma_service = MockAlmaService()

    update_func(system_identity, "aiekd-23382", "77777", alma_service)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Graz University of Technology.
#
# invenio-workflows-tugraz is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Alembic create tables for theses workflow."""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "eca8ae6a6bc1"
down_revision = None
branch_labels = ()
depends_on = "02845ec6893c"


def upgrade() -> None:
    """Upgrade database."""
    op.add_column(
        "workflows_theses",
        sa.Column("imported_in_repo", sa.Boolean(), default=False),
    )
    op.add_column(
        "workflows_theses",
        sa.Column("updated_in_repo", sa.Boolean(), default=False),
    )
    op.drop_column("workflows_theses", "ready_to_archive_in_cms")
    op.drop_column("workflows_theses", "ready_to_create_in_alma")
    op.drop_column("workflows_theses", "ready_to_update_in_repo")
    op.drop_column("workflows_theses", "ready_to_publish_in_cms")


def downgrade() -> None:
    """Downgrade database."""
    op.add_column(
        "workflows_theses",
        sa.Column("ready_to_archive_in_cms", sa.Boolean(), default=False),
    )
    op.add_column(
        "workflows_theses",
        sa.Column("ready_to_create_in_alma", sa.Boolean(), default=False),
    )
    op.add_column(
        "workflows_theses",
        sa.Column("ready_to_update_in_repo", sa.Boolean(), default=False),
    )
    op.add_column(
        "workflows_theses",
        sa.Column("ready_to_publish_in_cms", sa.Boolean(), default=False),
    )
    op.drop_column("workflows_theses", "imported_in_repo")
    op.drop_column("workflows_theses", "updated_in_repo")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Graz University of Technology.
#
# invenio-workflows-tugraz is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Types for workflows."""


from dataclasses import dataclass


@dataclass(frozen=True)
class Color:
    """The class is for the output color management."""

    neutral = "white"
    error = "red"
    warning = "yellow"
    abort = "magenta"
    success = "green"
    alternate = ("blue", "cyan")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2024 Graz University of Technology.
#
# invenio-workflows-tugraz is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Initialize the package."""

from flask import Flask

from . import config
from .theses import WorkflowThesesService, WorkflowThesesServiceConfig


class InvenioWorkflowsTugraz:
    """invenio-workflows-tugraz extension."""

    def __init__(self, app: Flask = None) -> None:
        """Extension initialization."""
        if app:
            self.init_app(app)

    def init_app(self, app: Flask) -> None:
        """Flask application initialization."""
        self.init_config(app)
        self.init_services(app)
        app.extensions["invenio-workflows-tugraz"] = self

    def init_config(self, app: Flask) -> None:
        """Initialize configuration."""
        for k in dir(config):
            attr = getattr(config, k)

            if k == "WORKFLOW_ALMA_REPOSITORY_RECORDS_UPDATE_AGGREGATOR":
                app.config["ALMA_REPOSITORY_RECORDS_UPDATE_AGGREGATOR"] = attr

            elif k == "WORKFLOW_ALMA_REPOSITORY_RECORDS_UPDATE_FUNC":
                app.config["ALMA_REPOSITORY_RECORDS_UPDATE_FUNC"] = attr

            elif k == "WORKFLOW_ALMA_ALMA_RECORDS_CREATE_AGGREGATOR":
                app.config["ALMA_ALMA_RECORDS_CREATE_AGGREGATOR"] = attr

            elif k == "WORKFLOW_ALMA_ALMA_RECORDS_CREATE_FUNC":
                app.config["ALMA_ALMA_RECORDS_CREATE_FUNC"] = attr

            elif k == "WORKFLOW_CAMPUSONLINE_THESES_FILTER":
                app.config["CAMPUSONLINE_THESES_FILTER"] = attr

            elif k == "WORKFLOW_CAMPUSONLINE_IMPORT_FUNC":
                app.config["CAMPUSONLINE_IMPORT_FUNC"] = attr

            elif k == "WORKFLOW_CAMPUSONLINE_DUPLICATE_FUNC":
                app.config["CAMPUSONLINE_DUPLICATE_FUNC"] = attr

            elif k.startswith("WORKFLOWS_TUGRAZ_"):
                app.config.setdefault(k, attr)

    def init_services(self, app: Flask) -> None:
        """Init services."""
        theses_config = WorkflowThesesServiceConfig.build(app)
        self.theses_service = WorkflowThesesService(config=theses_config)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 Graz University of Technology.
#
# invenio-workflows-tugraz is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.


"""Decorator functions for theses views."""

from collections.abc import Callable
from functools import wraps
from typing import Any

from flask import abort, g
from invenio_pidstore.errors import PIDDoesNotExistError
from invenio_records_marc21.proxies import current_records_marc21
from invenio_search import RecordsSearch
from invenio_search.engine import dsl
from sqlalchemy.exc import NoResultFound


def pass_record_from_pid(f: Callable) -> Callable:
    """Decorate a view to pass the record from a pid."""

    @wraps(f)
    def view(*_: dict, **kwargs: dict) -> Any:  # noqa: ANN401
        pid_value = kwargs.get("pid_value")

        search = RecordsSearch(index="marc21records")
        query = {
            "filter": [
                {"match_all": {}},
                {"match_phrase": {"metadata.fields.995.subfields.a": pid_value}},
                {"match_phrase": {"metadata.fields.995.subfields.i": "TUGRAZonline"}},
            ],
        }

        search.query = dsl.Q("bool", **query)
        search = search.params(size=1)
        result = search.execute()
        if len(result["hits"]["hits"]) == 0:
            abort(423)
        hits = result["hits"]["hits"][0]["_source"]
        hits = hits.to_dict()

        try:
            record = current_records_marc21.records_service.read_draft(
                id_=hits["id"],
                identity=g.identity,
            )
        except (PIDDoesNotExistError, NoResultFound):
            record = current_records_marc21.records_service.read(
                id_=hits["id"],
                identity=g.identity,
            )

        kwargs["record"] = record.to_dict()
        return f(**kwargs)

    return view



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2024 Graz University of Technology.
#
# invenio-workflows-tugraz is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Configuration file."""

from .openaccess import pure_import_func, pure_sieve_func
from .theses import (
    create_func,
    duplicate_func,
    import_from_alma_func,
    import_from_cms_func,
    theses_create_aggregator,
    theses_filter,
    theses_update_aggregator,
    update_func,
)

WORKFLOW_ALMA_REPOSITORY_RECORDS_IMPORT_FUNC = import_from_alma_func
""""""

WORKFLOW_ALMA_REPOSITORY_RECORDS_UPDATE_AGGREGATOR = theses_update_aggregator
""""""

WORKFLOW_ALMA_REPOSITORY_RECORDS_UPDATE_FUNC = update_func
""""""

WORKFLOW_ALMA_ALMA_RECORDS_CREATE_FUNC = create_func
""""""

WORKFLOW_ALMA_ALMA_RECORDS_CREATE_AGGREGATOR = theses_create_aggregator
""""""

WORKFLOW_CAMPUSONLINE_THESES_FILTER = theses_filter()
""""""

WORKFLOW_CAMPUSONLINE_IMPORT_FUNC = import_from_cms_func
""""""

WORKFLOW_CAMPUSONLINE_DUPLICATE_FUNC = duplicate_func
""""""

WORKFLOW_PURE_IMPORT_FUNC = pure_import_func
"""See corresponding varaible in invenio-pure."""

WORKFLOW_PURE_SIEVE_FUNC = pure_sieve_func



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Graz University of Technology.
#
# invenio-workflows-tugraz is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Proxies."""

from flask import current_app
from werkzeug.local import LocalProxy

current_workflows_tugraz = LocalProxy(
    lambda: current_app.extensions["invenio-workflows-tugraz"],
)
"""Helper proxy to get the current workflows tugraz extension."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Graz University of Technology.
#
# invenio-workflows-tugraz is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Alembic create tables for theses workflow."""

import sqlalchemy as sa
import sqlalchemy_utils
from alembic import op

# revision identifiers, used by Alembic.
revision = "2b99bb26b381"
down_revision = None
branch_labels = ()
depends_on = "9d446c6a77e2"


def upgrade() -> None:
    """Upgrade database."""
    op.create_table(
        "workflows_theses",
        sa.Column(
            "id",
            sqlalchemy_utils.types.uuid.UUIDType(),
            nullable=False,
        ),
        sa.Column(
            "cms_id",
            sa.Integer(),
            autoincrement=False,
            nullable=False,
        ),
        sa.Column(
            "archived",
            sa.Boolean(),
            default=False,
        ),
        sa.Column(
            "published",
            sa.Boolean(),
            default=False,
        ),
        sa.Column(
            "ready_to_archive",
            sa.Boolean(),
            default=False,
        ),
        sa.Column(
            "ready_to_publish",
            sa.Boolean(),
            default=False,
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_workflows_theses")),
    )


def downgrade() -> None:
    """Downgrade database."""
    op.drop_table("workflows_theses")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Graz University of Technology.
#
# invenio-workflows-tugraz is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Alembic create tables for theses workflow."""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "02845ec6893c"
down_revision = None
branch_labels = ()
depends_on = "2b99bb26b381"


def upgrade() -> None:
    """Upgrade database."""
    op.add_column(
        "workflows_theses",
        sa.Column("archived_in_cms", sa.Boolean(), default=False),
    )
    op.add_column(
        "workflows_theses",
        sa.Column("created_in_alma", sa.Boolean(), default=False),
    )
    op.add_column(
        "workflows_theses",
        sa.Column("published_in_cms", sa.Boolean(), default=False),
    )
    op.add_column(
        "workflows_theses",
        sa.Column("ready_to_archive_in_cms", sa.Boolean(), default=False),
    )
    op.add_column(
        "workflows_theses",
        sa.Column("ready_to_create_in_alma", sa.Boolean(), default=False),
    )

    op.add_column(
        "workflows_theses",
        sa.Column("ready_to_update_in_repo", sa.Boolean(), default=False),
    )

    op.add_column(
        "workflows_theses",
        sa.Column("ready_to_publish_in_cms", sa.Boolean(), default=False),
    )
    op.drop_column("workflows_theses", "archived")
    op.drop_column("workflows_theses", "published")
    op.drop_column("workflows_theses", "ready_to_archive")
    op.drop_column("workflows_theses", "ready_to_publish")


def downgrade() -> None:
    """Downgrade database."""
    op.drop_column("workflows_theses", "archived_in_cms")
    op.drop_column("workflows_theses", "created_in_alma")
    op.drop_column("workflows_theses", "published_in_cms")
    op.drop_column("workflows_theses", "ready_to_archive_in_cms")
    op.drop_column("workflows_theses", "ready_to_create_in_alma")
    op.drop_column("workflows_theses", "ready_to_update_in_repo")
    op.drop_column("workflows_theses", "ready_to_publish_in_cms")

    op.add_column(
        "workflows_theses",
        sa.Column("archived", sa.Boolean(), default=False),
    )
    op.add_column(
        "workflows_theses",
        sa.Column("published", sa.Boolean(), default=False),
    )
    op.add_column(
        "workflows_theses",
        sa.Column("ready_to_archive", sa.Boolean(), default=False),
    )
    op.add_column(
        "workflows_theses",
        sa.Column("ready_to_publish", sa.Boolean(), default=False),
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Graz University of Technology.
#
# invenio-workflows-tugraz is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Alembic create tables for theses workflow."""

import sqlalchemy as sa
from alembic import op
from sqlalchemy_utils import UUIDType

# revision identifiers, used by Alembic.
revision = "937ba80502f3"
down_revision = None
branch_labels = ()
depends_on = "02845ec6893c"


def upgrade() -> None:
    """Upgrade database."""
    op.alter_column(
        "workflows_theses",
        "id",
        new_column_name="pid",
        existing_type=UUIDType(),
        type_=sa.String(255),
    )


def downgrade() -> None:
    """Downgrade database."""
    op.alter_column(
        "workflows_theses",
        "pid",
        new_column_name="id",
        existing_type=sa.String(255),
        type_=sa.UUIDType(),
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Graz University of Technology.
#
# invenio-workflows-tugraz is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Config for theses workflow."""

from invenio_records_resources.services.base.config import (
    ConfiguratorMixin,
    ServiceConfig,
)

from .api import WorkflowTheses


class WorkflowThesesServiceConfig(ServiceConfig, ConfiguratorMixin):
    """Workflow theses service config."""

    theses_cls = WorkflowTheses



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2024 Graz University of Technology.
#
# invenio-workflows-tugraz is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Theses Workflows."""


from pathlib import Path
from typing import NamedTuple, Union

from flask_principal import Identity
from invenio_access.permissions import system_identity
from invenio_alma import AlmaRESTService, AlmaSRUService
from invenio_alma.services import AlmaAPIError, AlmaRESTError
from invenio_alma.utils import is_duplicate_in_alma, validate_date
from invenio_campusonline import CampusOnlineRESTService
from invenio_campusonline.records.models import CampusOnlineRESTError
from invenio_campusonline.types import CampusOnlineID, ThesesFilter
from invenio_campusonline.utils import extract_embargo_range
from invenio_pidstore.errors import PIDDoesNotExistError
from invenio_records_marc21 import (
    DuplicateRecordError,
    Marc21Metadata,
    MarcDraftProvider,
    check_about_duplicate,
    convert_json_to_marc21xml,
    create_record,
    current_records_marc21,
)
from invenio_records_marc21.services.record.types import ACNumber
from invenio_records_resources.services.records.results import RecordItem
from marshmallow.exceptions import ValidationError
from sqlalchemy.orm.exc import NoResultFound, StaleDataError

from ..proxies import current_workflows_tugraz
from .convert import CampusOnlineToMarc21
from .types import CampusOnlineId

error_record = NamedTuple("ErrorRecord", ["id"])


@check_about_duplicate.register
def _(value: CampusOnlineId) -> None:
    """Check about double campus online id."""
    check_about_duplicate(str(value), value.category)


def theses_filter() -> ThesesFilter:
    """Return a ThesesFilter object for open records.

    FILTER: xml filter to get open records
    return ThesesFilter
    """
    start_time_tag = "<bas:from>2022-11-17T00:01:00+00:00</bas:from>"
    filter_ = f"""
        <bas:thesesType>ALL</bas:thesesType>
        <bas:state name="IFG" negate="false">{start_time_tag}</bas:state>
        <bas:state name="PUBLISHABLE" negate="false"></bas:state>
        <bas:state name="ARCH" negate="true"></bas:state>
        <bas:state name="PUB" negate="true"></bas:state>
    """

    return ThesesFilter(filter_)


def theses_create_aggregator() -> list[tuple[str, str]]:
    """Return list of marc21,cmsid tuple which should be created in alma."""
    theses_service = current_workflows_tugraz.theses_service
    return theses_service.get_ready_to(system_identity, state="create_in_alma")


def theses_update_aggregator() -> list[tuple[str, str]]:
    """Return a list of tuple(marc21, cms_id) which should be updated in repo."""
    theses_service = current_workflows_tugraz.theses_service
    return theses_service.get_ready_to(system_identity, state="update_in_repo")


def import_from_alma_func(
    identity: Identity,
    ac_number: str,
    file_path: str,
    access: str,
    embargo: Union[str, None] = None,  # change: str|None after end python3.9 support
    marcid: Union[str, None] = None,
    alma_service: AlmaSRUService = None,
    **_: dict,
) -> None:
    """Process a single import by cli of a alma record by ac number.

    Embargo has to be YYYY-MM-DD
    """
    if not alma_service:
        msg = "ERROR: alma_service for import_from_alma_func not set."
        raise RuntimeError(msg)
    marc21_service = current_records_marc21.records_service

    if marcid:
        MarcDraftProvider.predefined_pid_value = marcid

    if embargo and not validate_date(embargo):
        msg = (f"NotValidEmbargo search_value: {ac_number}, embargo: {embargo}",)
        raise RuntimeError(msg)

    try:
        check_about_duplicate(ACNumber(ac_number))
    except DuplicateRecordError as error:
        raise RuntimeError(str(error)) from error

    try:
        metadata = alma_service.get_record(ac_number)[0]
    except AlmaRESTError as error:
        msg = f"ERROR: alma rest search_value: {ac_number}, error: {error}"
        raise RuntimeError(msg) from error

    marc21_record = Marc21Metadata(metadata=metadata)

    data = marc21_record.json
    data["access"] = {
        "record": "public",
        "files": "public" if access == "public" else "restricted",
    }

    if not Path(file_path).is_file():
        msg = f"ERROR: FileNotFoundError search_value: {ac_number}, file_path: {file_path}"  # noqa: E501
        raise RuntimeError(msg)

    if embargo:
        data["access"]["embargo"] = {
            "until": embargo,
            "active": True,
            "reason": None,
        }

    try:
        record = create_record(marc21_service, data, [file_path], identity)
    except StaleDataError as error:
        msg = f"ERROR: StaleDataError search_value: {ac_number}"
        raise RuntimeError(msg) from error
    except ValidationError as error:
        msg = f"ValidationError   search_value: {ac_number}, error: {error}"
        raise RuntimeError(msg) from error

    return record


def import_from_cms_func(
    identity: Identity,
    cms_id: CampusOnlineID,
    cms_service: CampusOnlineRESTService,
) -> RecordItem:
    """Import the record into the repository from campusonline."""
    marc21_service = current_records_marc21.records_service
    theses_service = current_workflows_tugraz.theses_service

    try:
        check_about_duplicate(CampusOnlineId(cms_id))
    except DuplicateRecordError as error:
        raise RuntimeError(str(error)) from error

    try:
        thesis = cms_service.get_metadata(identity, cms_id)
        file_path = cms_service.download_file(identity, cms_id)
    except CampusOnlineRESTError as error:
        raise RuntimeError(str(error)) from error

    marc21_record = Marc21Metadata()
    converter = CampusOnlineToMarc21(marc21_record)
    converter.convert(thesis, marc21_record)

    data = marc21_record.json
    data["access"] = {
        "record": "restricted",
        "files": "restricted",
    }

    if bool(_ := extract_embargo_range(thesis)):
        # the embargo end date from tugonline is ignored on purpose and set to a
        # infinity value to express that the enddate will not be reached!
        # the requirement forces a manual removal of the embargo.
        data["access"]["embargo"] = {
            "until": "9999-12-12",
            "active": True,
            "reason": None,
        }

    try:
        record = create_record(
            marc21_service,
            data,
            [file_path],
            identity,
            do_publish=False,
        )
    except StaleDataError as error:
        msg = f"ERROR: StaleDataError cms_id: {cms_id}"
        raise RuntimeError(msg) from error
    except ValidationError as error:
        msg = f"ValidationError cms_id: {cms_id}, error: {error}"
        raise RuntimeError(msg) from error

    theses_service.create(identity, record.id, cms_id)
    theses_service.set_state(identity, id_=record.id, state="imported_in_repo")

    return record


def create_func(
    identity: Identity,
    marc_id: str,
    cms_id: str,
    alma_service: AlmaRESTService,
) -> None:
    """Create a record in alma.

    Normally - depending on the API_KEY - the record will be created in
    the Institution Zone (IZ).
    """
    marc21_service = current_records_marc21.records_service
    theses_service = current_workflows_tugraz.theses_service

    if is_duplicate_in_alma(cms_id):
        msg = f"WARNING: duplicate in alma cms_id: {cms_id}"
        raise RuntimeWarning(msg)

    try:
        record = marc21_service.read_draft(identity, marc_id)
    except (NoResultFound, PIDDoesNotExistError) as error:
        msg = f"ERROR: marc_id: {marc_id}, cms_id: {cms_id} not found in db"
        raise RuntimeError(msg) from error

    marc21_record_etree = convert_json_to_marc21xml(record.to_dict()["metadata"])

    try:
        alma_service.create_record(marc21_record_etree)
    except AlmaRESTError as error:
        msg = f"ERROR: alma rest error on marc_id: {marc_id}, cms_id: {cms_id}, error: {error}"  # noqa: E501
        raise RuntimeError(msg) from error

    theses_service.set_state(identity, id_=marc_id, state="created_in_alma")


def update_func(
    identity: Identity,
    marc_id: str,
    cms_id: str,
    alma_service: AlmaSRUService,
    *,
    update_access: bool = True,
) -> None:
    """Update the record by metadata from alma.

    :param bool update_access: normally true, but if updated by cli it would be
    nice to update records without worring to mess up the access.
    """
    marc21_service = current_records_marc21.records_service
    theses_service = current_workflows_tugraz.theses_service

    try:
        data = marc21_service.read_draft(id_=marc_id, identity=identity).data
    except (NoResultFound, PIDDoesNotExistError):
        try:
            # if this raises also the NoResultFound error it should break!
            data = marc21_service.read(id_=marc_id, identity=identity).data
        except (NoResultFound, PIDDoesNotExistError) as error:
            msg = f"ERROR: update record marc_id: {marc_id}, cms_id: {cms_id} not found in db"  # noqa: E501
            raise RuntimeError(msg) from error

    if update_access:
        db_marc21_record = Marc21Metadata(json=data["metadata"])

        # The existens of the "gesperrt" field will be checked from the
        # database metadata because the field could be removed by
        # accident. It would be a feature to open the files within the
        # repository by the metadata comming from alma but the risk of
        # exposing files without intention is to high.
        is_restricted = db_marc21_record.exists_field(
            category="971",
            ind1="7",
            ind2=" ",
            subf_code="a",
            subf_value="gesperrt",
        )
        data["access"]["record"] = "public"
        data["access"]["files"] = "restricted" if is_restricted else "public"

    try:
        alma_marc21_etree = alma_service.get_record(cms_id, "local_field_995")
    except (AlmaRESTError, AlmaAPIError) as error:
        msg = f"ERROR: alma rest marc_id: {marc_id}, cms_id: {cms_id}, error: {error}"
        raise RuntimeError(msg) from error

    alma_marc21_record = Marc21Metadata(metadata=alma_marc21_etree[0])

    # only update and publish records which are associated with
    # "verbund" and have therefore an AC* number. this also shows that
    # the record was viewed by a librarian
    ac_field = alma_marc21_record.get_value("009")
    if not ac_field.startswith("AC"):
        msg = f"marcid: {marc_id}, cms_id: {cms_id} not yet updated in alma"
        raise RuntimeError(msg)

    data["metadata"] = alma_marc21_record.json["metadata"]

    try:
        marc21_service.edit(id_=marc_id, identity=identity)
        marc21_service.update_draft(id_=marc_id, identity=identity, data=data)
        marc21_service.publish(id_=marc_id, identity=identity)
    except ValidationError as error:
        msg = f"ValidationError cms_id: {cms_id}, error: {error}"
        raise RuntimeError(msg) from error

    theses_service.set_state(identity, id_=marc_id, state="updated_in_repo")


def duplicate_func(cms_id: str) -> bool:
    """Check if the cms_id has already been imported."""
    try:
        check_about_duplicate(CampusOnlineId(cms_id))
        return False
    except DuplicateRecordError:
        return True
